import os

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as pkg_resources

from copyreg import constructor
from gc import callbacks
from notebookjs import execute_js
from .helpers import calculate_histograms, calculate_preds_histograms
from .reliabilitycurve import ReliabilityCurve
import numpy as np
# from callbacks import reliability_diagram, learned_reliability_diagram, filter_by_range, filter_by_feature_range
from .callbacks import get_reliability_curve, learned_reliability_diagram, confusion, get_table_average



class Calibrate:

    def __init__(self, data) -> None:

        ## calculating histograms
        self.histograms = calculate_histograms( data )

        ## user-generated data
        self.data = data

        ## models
        self.predictions = {}
        self.labels = {}
        self.models = []
        
        ## number of classes
        self.nClasses = 0

        ## created curves
        self.createdCurves = []

        ## loading vis lib
        self.vislib = None
        data_dir = os.path.join(os.path.dirname(__file__), "")
        data_path = os.path.join(data_dir, "vis/dist/calibration.js")
        with open(data_path, "r") as f:
            self.vislib = f.read()

    def add_model(self, predictions, labels, name):

        ## TODO: make sure all models have same number of classes

        ## appending model name
        self.models.append(name)

        ## appending model data
        self.labels[name] = labels
        self.predictions[name]  = predictions

        ## updating nclasses
        self.nClasses = predictions.shape[1]
        
    def visualize(self):

        ## setting callbacks
        callbacks = {
            'get_reliability_curve': self.get_reliability_curve,
            'get_learned_curve': self.get_learned_curve,
            'get_curve_instance_data': self.get_curve_instance_data,
            'clear_curves': self.clear_curves,
            'filter_by_pred_range': self.filter_by_pred_range
        }

        ## setting input data
        input_data = {
            'histograms': self.histograms,
            'models': self.models,
            'nclasses': self.nClasses
        }

        # Plotting the Radial Bar Chart
        execute_js(
            library_list=[self.vislib],
            main_function="calibration.renderCalibration", 
            data_dict=input_data,
            callbacks=callbacks )

    def get_preds_histogram( self, preds ):
        return calculate_preds_histograms(preds)

    def get_reliability_curve( self, event ):

        ## current model
        modelPredictions = self.predictions[event['currentmodel']]
        modelLabels = self.labels[event['currentmodel']]
        
        ## getting curve
        currentcurvedata, instancedata, confusionmatrix, preds, labels, allClassPreds, allClassLabels = get_reliability_curve( event, data=self.data, preds=modelPredictions, labels=modelLabels )
        
        ## saving instance data
        curve = ReliabilityCurve( 
            tableheader=instancedata['tableheader'], 
            tablebody=instancedata['tablebody'], 
            confusionMatrix=confusionmatrix, 
            preds=preds, 
            labels=labels,
            allClassPreds=allClassPreds,
            allClassLabels=allClassLabels)
        self.createdCurves.append(curve)

        ## getting curve
        return currentcurvedata

    def get_learned_curve( self, event ):

        selectedCurve = self.createdCurves[event['curveIndex']]
        learnedCurve = learned_reliability_diagram( selectedCurve.preds, selectedCurve.labels )

        return learnedCurve

    def get_curve_instance_data( self, event ):

        currentInstanceData = {
            'tableheader':  self.createdCurves[event['curveIndex']].tableheader,
            'tablebody': self.createdCurves[event['curveIndex']].tablebody,
            'tableaverages': get_table_average(self.createdCurves[event['curveIndex']].tablebody),
            'confusionmatrix': self.createdCurves[event['curveIndex']].confusionMatrix,
            'predshistogram': self.get_preds_histogram(self.createdCurves[event['curveIndex']].preds)
        }

        return currentInstanceData

    def filter_by_pred_range(self, event ):

        ## getting current curve
        currentCurve = self.createdCurves[0]

        ## setting conds
        conds = ( (currentCurve.preds >= event['start']) & (currentCurve.preds <= event['end']) )
        tablebody = np.array(currentCurve.tablebody)[conds]
        
        ## calculating filtered confusion matrix
        preds = currentCurve.allClassPreds[conds]
        labels = currentCurve.allClassLabels[conds]
        confusionMatrix = confusion(preds, labels)

        return {
            'tablebody': tablebody.tolist(),
            'tableheader': currentCurve.tableheader,
            'tableaverages': get_table_average( tablebody ),
            'confusionmatrix': confusionMatrix
        }

    def clear_curves(self, event):

        nCurves = len(self.createdCurves)
        self.createdCurves = []

        return {'ncurves':  nCurves}




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pycalibrate",
    version="0.0.3",
    author="Peter Xenopoulos, Joao Rulff, Brian Barr, Luis Gustavo Nonato, Claudio Silva",
    author_email="xenopoulos@nyu.edu",
    description="pycalibrate. A tool to assess classifier calibration in Jupyter Notebooks",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/VIDA-NYU/pycalibrate',
    packages=find_packages(exclude=['node_modules']),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'numpy',
        'scikit-learn',
        'notebook',
        'notebookjs',
        'interpret',
        'pandas'
    ]
)



from sklearn.calibration import calibration_curve
from sklearn.metrics import confusion_matrix
from interpret.glassbox import ExplainableBoostingClassifier
import numpy as np
import pandas as pd


def get_table_average( tablebody  ):

    return  np.around( np.array(tablebody).mean(axis=0) , decimals=2).tolist()

def get_reliability_curve( filters, data, preds, labels ):

    filteredPreds = preds
    filteredLabels = labels
    filteredData = data

    ## Filtering by features
    for key in filters['featurefilters']:

        ## setting boolean conditions
        currentconds = ( (filteredData[filters['featurefilters'][key]['name']] >= filters['featurefilters'][key]['start']) & (filteredData[filters['featurefilters'][key]['name']] <= filters['featurefilters'][key]['end']) )
        
        ## filtering dataframes
        filteredData = filteredData[ currentconds ]
        filteredPreds = filteredPreds[ currentconds ] 
        filteredLabels = filteredLabels[ currentconds ]


    ## saving all classes predictions
    allClassPreds = filteredPreds
    allClassLabels = filteredLabels

    ## confusion matrix
    confusionMatrix = confusion(filteredPreds, filteredLabels)

    ## filtering to current class index
    filteredPreds = filteredPreds[:, filters['selectedclass']]
    filteredLabels = filteredLabels[:, filters['selectedclass']]

    # new reliability curve
    chartData = reliability_diagram( preds=filteredPreds, labels=filteredLabels, bins=filters['nbins'] )

            
    return {'reliabilitychart': chartData }, {
            'tableheader': filteredData.columns.tolist(),
            'tablebody': np.around(filteredData.values, decimals=2).tolist(),
            # 'classifications': filteredLabels.tolist()
        }, confusionMatrix, filteredPreds, filteredLabels, allClassPreds, allClassLabels


def reliability_diagram( preds, labels, bins ):

    '''
        n = number of samples
        k = number of classseex

        preds - n x k numpy array of predicted probabilities
        labels - n x k numpy array of one-hot encoded labels
        class_index - integer of what class to consider
        bins - integer for number of bins used to make curve
    '''

    acc, conf = calibration_curve(labels, preds, n_bins=bins, strategy="uniform")
    chartData = [ { 'x': conf[i], 'y': acc[i] }  for i in range(acc.shape[0]) ]
    return chartData

def confusion( preds, labels ):
    
    y_pred = np.argmax(preds, axis=1)
    y_true = np.argmax(labels, axis=1)
    
    return confusion_matrix(y_true, y_pred)

def learned_reliability_diagram(preds, labels, bins=256, random_state=42):

    '''
        n = number of samples
        k = number of classes

        preds - n x k numpy array of predicted probabilities
        labels - n x k numpy array of one-hot encoded labels
        class_index - integer of what class to consider
        bins - integer for number of bins in EBM model
        random_state - for reproducibility
    '''

    ebm = ExplainableBoostingClassifier(random_state=random_state, binning="uniform", max_bins=bins)
    ebm.fit(preds.reshape(-1,1), labels.reshape(-1,1))
    conf = np.linspace(0,1,num=100)
    acc = ebm.predict_proba(conf.reshape(-1,1))[:,1]

    chartData = [ { 'x': conf[i], 'y': acc[i] }  for i in range(acc.shape[0])]
    return {'learnedcurve': chartData }


import numpy as np

def calculate_preds_histograms( preds ):
    currentHistogram = np.histogram(preds, bins=20)
    histvalues = currentHistogram[0].tolist()
    histbounds = currentHistogram[1].tolist()
    currentFeature = { 'values': histvalues, 'bounds': histbounds }
    return currentFeature

def calculate_histograms( data ):

    histograms = []
    for column in data.columns:
        
        currentHistogram = np.histogram(data[column].values, bins=20)
        histvalues = currentHistogram[0].tolist()
        histbounds = currentHistogram[1].tolist()
        currentFeature = { 'name': column, 'values': histvalues, 'bounds': histbounds }
        histograms.append(currentFeature)
        
    return histograms


class ReliabilityCurve:

    def __init__(self, tableheader, tablebody, confusionMatrix, preds, labels, allClassPreds, allClassLabels ):

        self.tableheader = tableheader
        self.tablebody = tablebody

        ## saving confusion matrix
        self.confusionMatrix = confusionMatrix

        self.preds = preds
        self.labels = labels


        ## TODO: Remove after the deadline. I added this to calculate the confusion matrix on brush faster. 
        ## However, we can just do a count of the remaining samples after filtering.
        self.allClassPreds = allClassPreds
        self.allClassLabels = allClassLabels

        # self.nBins = 0
        # self.model = 0
